import numpy as np
import pyRBM.Simulation.State as State

class Solver:
    def __init__(self, locations, rules, matched_indices, model_state:State.ModelState, use_cached_propensities:bool = True, no_rules_behaviour:str = "step", propensity_update_dict:dict = {}):
        self.locations = locations
        self.rules = rules
        self.matched_indices = matched_indices
        self.model_state = model_state

        self.use_cached_propensities = use_cached_propensities

        self.last_rule_index_set = None

        self.propensities = {}
        if self.use_cached_propensities:
            self.total_propensity = 0
        self.propensity_update_dict = propensity_update_dict
        # Either step or exit
        assert (no_rules_behaviour in ["step", "exit"])
        self.no_rules_behaviour = no_rules_behaviour
        self.default_step = 1

    def simulateOneStep(self):
        raise(TypeError("Abstract class Solver, please use a concrete implementation."))
    
    # rules_and_matched_indices is used to determine which propensities to recompute, if None is provided this is all propensities.
    # returns total propensity
    def updateGivenPropensities(self, rules_and_matched_indices = None):
        model_state_values = list(self.model_state.returnModelClassesValues())

        if rules_and_matched_indices is None:
            for rule_i in range(len(self.matched_indices)):
                rule = self.rules[rule_i]
                for index_set_i in range(len(self.matched_indices[rule_i])):
                    new_propensity = rule.returnPropensity(np.take(self.locations, self.matched_indices[rule_i][index_set_i]),
                                                   model_state_values)
                    if self.use_cached_propensities:
                        self.total_propensity += (new_propensity - self.propensities.get(f"{rule_i} {index_set_i}", 0.0))
                    self.propensities[f"{rule_i} {index_set_i}"] = new_propensity
        else:
            for rule_i, index_set_i in rules_and_matched_indices:
                rule = self.rules[rule_i]
                new_propensity = rule.returnPropensity(np.take(self.locations, self.matched_indices[rule_i][index_set_i]),
                                                                                     model_state_values)
                if self.use_cached_propensities:
                    self.total_propensity += (new_propensity - self.propensities.get(f"{rule_i} {index_set_i}", 0.0))
                self.propensities[f"{rule_i} {index_set_i}"] = new_propensity

    def performPropensityUpdates(self):
        if not self.last_rule_index_set is None:
            rule_prop_update_set = self.propensity_update_dict[self.last_rule_index_set]
            changed_base_model_vars = self.model_state.returnChangedVars()
            if not len(changed_base_model_vars) == 0:
                rule_prop_update_set = rule_prop_update_set.copy()
                for changed_model_var in changed_base_model_vars:
                    rule_prop_update_set.update(self.propensity_update_dict.get(changed_model_var, set()))
                    if changed_model_var == "model_month_jan":
                        print(self.propensity_update_dict.get(changed_model_var, set()))
            self.updateGivenPropensities(rule_prop_update_set)
        else:
            self.updateGivenPropensities()
    
    def reset(self):
        self.propensities = {}
        self.last_rule_index_set = None

    # CACHE TOTAL PROPENSITY AND UPDATE USING DIFF BETWEEN OLD AND NEW CACHE VALUES.
    def returnTotalPropensity(self):
        if not self.use_cached_propensities:
            return np.sum(list(self.propensities.values()))
        else:
            return self.total_propensity
class GillespieSolver(Solver):
    def __init__(self, locations, rules, matched_indices, model_state:State.ModelState, use_cached_propensities:bool = True, no_rules_behaviour:str = "step", propensity_update_dict:dict = {}):
        super().__init__(locations, rules, matched_indices, model_state, use_cached_propensities, no_rules_behaviour, propensity_update_dict)
    
    def simulateOneStep(self, current_time):
        self.performPropensityUpdates()

        total_propensity = self.returnTotalPropensity()
        
        if total_propensity <= 0:
            if self.no_rules_behaviour == "end":
                print("Finishing model simulation early.\n No rules left to trigger - all rules have 0 propensity.")
                return
            elif self.no_rules_behaviour == "step":
                print(f"Stepping {self.default_step} ahead.\n No rules left to trigger - all rules have 0 propensity.")
                return current_time + self.default_step
        
        # Generate 0 to 1
        # Random rule
        u1, r2 = np.random.random_sample(2)
        u2 = -np.log(r2)*(1/total_propensity)
        # Random time
        cumulative_prop = 0
        selected_rule = None
        for rule_loc_key in list(self.propensities.keys()):
            cumulative_prop += self.propensities[rule_loc_key]
            if cumulative_prop > u1*total_propensity:
                selected_rule = rule_loc_key
                break
        # Only set the last rule used when using the caching for propensities.
        if self.use_cached_propensities:
            self.last_rule_index_set = selected_rule
        selected_rule, selected_locations = selected_rule.split(" ")
        assert (self.rules[int(selected_rule)].triggerAttemptedRuleChange(np.take(self.locations, self.matched_indices[int(selected_rule)][int(selected_locations)])))
        return current_time + u2

class TauLeapingGillespieSolver(Solver):
    def __init__(self, locations, rules, matched_indices, model_state, use_cached_propensities:bool = True, no_rules_behaviour:str = "step"):
        super().__init__(locations, rules, matched_indices, model_state, use_cached_propensities, no_rules_behaviour)
    
    def simulateOneStep(self, current_time):
        total_propensity = 0
        propensities = []


        for rule_i in range(len(self.matched_indices)):
            rule = self.rules[rule_i]
            for index_set_I in range(len(self.matched_indices[rule_i])):
                propensity = rule.returnPropensity(np.take(self.locations, self.matched_indices[rule_i][index_set_I]), 
                                                   list(self.model_state.returnModelClassesValues()))
                total_propensity+= propensity
                # First element the propensity, second element the rule index, index element pair.
                propensities.append([propensity, [rule_i, index_set_I]])
                
        if total_propensity <= 0:
            if self.no_rules_behaviour == "end":
                print("Finishing model simulation early.\n No rules left to trigger - all rules have 0 propensity.")
                return
            elif self.no_rules_behaviour == "step":
                print(f"Stepping {self.default_step} ahead.\n No rules left to trigger - all rules have 0 propensity.")
                return current_time + self.default_step
        # Generate 0 to 1
        # Random rule
        u1, r2 = np.random.random_sample(2)
        u2 = -np.log(r2)*(1/total_propensity)
        # Random time
        cumulative_prop = 0
        selected_rule = None
        for rule_index, propensity in enumerate(propensities):
            if cumulative_prop+propensity[0] > u1*total_propensity:
                selected_rule = rule_index
                break
            else:
                cumulative_prop += propensity[0]
        selected_rule, selected_locations = propensities[selected_rule][1]
        assert (self.rules[selected_rule].triggerAttemptedRuleChange(np.take(self.locations, self.matched_indices[selected_rule][selected_locations])))
        return current_time + u2



import matplotlib.pyplot as plt
import numpy as np

class Trajectory:
    def __init__(self, locations) -> None:
        self.timestamps = {location_index:[0] for location_index, _ in enumerate(locations)}
        self.trajectory_location_values = {location_index:[location.class_values] for location_index, location in enumerate(locations)}

        self.last_time = 0
        self.last_location_index = None

        self.location_labels = {location_index:location.label_mapping for location_index, location in enumerate(locations)}
        self.location_names = {location_index:location.name for location_index, location in enumerate(locations)}

    def addEntry(self, time, location_values, location_index):
        if location_index != self.last_location_index:
            # Add last known time and value 
            self.trajectory_location_values[location_index].append(self.trajectory_location_values[location_index]
                                                                   [len(self.trajectory_location_values[location_index])-1])
            self.timestamps[location_index].append(self.last_time)

        self.trajectory_location_values[location_index].append(location_values)
        self.timestamps[location_index].append(time)

        self.last_time = time
        self.last_location_index = location_index

    def plotAllClassesOverTime(self, location_index, figure_position = "center left"):
        # ALLOW NONE AS ENTRY
        class_values = np.array(self.trajectory_location_values[location_index])
        for class_i in range(len(class_values[0])):
            plt.plot(self.timestamps[location_index], class_values[:,class_i])
        plt.legend([self.location_labels[location_index][str(i)].replace("_", " ") for i in range(len(self.location_labels[location_index]))], loc=figure_position)
        plt.title(f"Classes over time for {self.location_names[location_index]}")
        plt.show()


import numpy as np
import pyRBM.Build.Rules as Rules


class SingleLocationRule(Rules.Rule):
       def __init__(self, target, propensity, stoichiomety,
                    propensity_classes, stoichiometry_classes,
                    rule_name = "SINGLE LOCATION RULE"):
              super().__init__(rule_name, [target])
              self.addLinearStoichiomety([0], [stoichiomety], [stoichiometry_classes])
              self.addSimplePropensityFunction([0], [propensity], [propensity_classes])

class SingleLocationProductionRule(SingleLocationRule):
       def __init__(self, target, reactant_classes, reactant_amount, 
                    product_classes, product_amount,
                    propensity, propensity_classes, 
                    rule_name="SINGLE LOCATION PRODUCTION RULE"):
            
            reactants_len = len(reactant_classes)
            product_len = len(product_classes)
            stoichiometry = np.zeros(reactants_len+product_len)
            for i in range(reactants_len):
                stoichiometry[i] = -reactant_amount[i]
            for i  in range(product_len):
                stoichiometry[i+reactants_len] = product_amount[i]
            stoichiometry_classes = reactant_classes+product_classes

            super().__init__(target, propensity, stoichiometry, propensity_classes, stoichiometry_classes, rule_name)

class TransportRule(Rules.Rule):
       def __init__(self, source, target, transport_class,
                          propensities, transport_amount, propensity_classes,
                          rule_name = "TRANSPORT RULE"):
              super().__init__(rule_name, [source, target])

              assert(len(propensities) == 2)
              assert(len(propensity_classes) == 2)

              source_stochiometry = np.zeros(1)
              source_stochiometry[0] = -transport_amount
              target_stochiometry = np.zeros(1)
              target_stochiometry[0] = transport_amount

              self.addLinearStoichiomety([0, 1], [source_stochiometry, target_stochiometry], [[transport_class], [transport_class]])
              self.addSimplePropensityFunction([0, 1], propensities, propensity_classes)

class ExitEntranceRule(Rules.Rule):
       def __init__(self, target, transport_class,
                    transport_amount, propensity, propensity_classes,
                    rule_name = "EXIT/ENTRANCE RULE"):
            super().__init__(rule_name, [target])

            self.addLinearStoichiomety([0], [[transport_amount]], [[transport_class]])
            self.addSimplePropensityFunction([0], [propensity], [propensity_classes])


class Location:
    def __init__ (self, index, name, lat, long, loc_type, label_mapping, initial_class_values, location_constants):
        self.index = index
        self.name = name
        self.lat = lat
        self.long = long
        self.loc_type = loc_type
        self.label_mapping = label_mapping

        self.initial_class_values = initial_class_values
        self.class_values = initial_class_values
        self.location_constants = location_constants
    
    def updateCompartmentValues(self, new_values):
        self.class_values = new_values
    
    def reset(self):
        self.class_values = self.initial_class_values



import numpy as np

class Location:
    def __init__(self, lat:float, long:float, name:str, loc_type:str, constants = None, loc_prefix:str="loc_"):
        self.lat = lat
        self.long = long
        self.name = name
        self.class_labels = set([])
        self.loc_type = loc_type
        self.loc_prefix = loc_prefix

        # Constants will not be saved at the moment and will be replaced in rule matching.
        
        self.location_constants = None
        if isinstance(constants, dict):
            self.addAndSetConstants(constants)
        elif isinstance(constants, list):
            self.addConstants(constants)

        self.locations_variables = {}

        self.inital_conditions_dict = None

    def addConstants(self, constants:list):
        if constants is not None:
            if self.location_constants is None:
                self.location_constants = {}
            self.location_constants = self.location_constants|{self.loc_prefix+constant:None for constant in constants}

    def checkConditionDict(self, inital_conditions_dict:dict):
        for condition_class in inital_conditions_dict:
            if condition_class not in self.class_labels:
                raise(ValueError(f"Initial condition class {condition_class} not present at location {self.name}."))
            
    def checkConstantsDefined(self):
        if not self.location_constants is None:
            for constant in list(self.location_constants.keys()):
                if self.location_constants[constant] is None:
                    raise(ValueError(f"Location constant {constant} not defined at location {self.name}."))
    
    def setInitialConditions(self, inital_conditions_dict:dict):
        # check all classes are defined
        self.checkConditionDict(inital_conditions_dict)
        self.inital_conditions_dict = inital_conditions_dict

    def setDistances(self, distances:list):
        self.addConstants([f"distance_{i}" for i in range(len(distances))])
        self.setConstants({f"distance_{i}":distance for i, distance in enumerate(distances)})
    
    def setConstants(self, constants_dict:dict):
        current_constant_keys = list(self.location_constants.keys())
        for entered_constant in list(constants_dict.keys()):
            if self.loc_prefix+entered_constant in current_constant_keys:
                self.location_constants[self.loc_prefix+entered_constant] = constants_dict[entered_constant]
            else:
                raise(ValueError(f"Provided constant {entered_constant}, does not exist at current location {self.name}\n Defined location constants: {str(self.location_constants)}"))
            
    def addAndSetConstants(self, constants_dict:dict):
        self.addConstants(list(constants_dict.keys()))
        self.setConstants(constants_dict)
    def returnConstantNames(self):
        if self.location_constants is not None:
            return list(self.location_constants.keys())
        else:
            return []

    def returnDictDescription(self):
        #For ease of use and to ensure compartment label order is invariant to order construction functions are called.
        self.checkConstantsDefined()
        ordered_class_labels = sorted(self.class_labels)
        class_label_mapping = dict((str(i),x) for i,x in enumerate(ordered_class_labels))
        initial_conds = list(np.zeros(len(self.class_labels)))

        if not self.inital_conditions_dict is None:
            self.checkConditionDict(self.inital_conditions_dict)
            initial_conds_keys = list(self.inital_conditions_dict.keys())
            for index, class_label in enumerate(ordered_class_labels):
                if class_label in initial_conds_keys:
                    initial_conds[index] = self.inital_conditions_dict[class_label]
        
        if self.location_constants is None:
            self.location_constants = {}

        return {"location_name":self.name, "lat" : self.lat, "long":self.long,
                "label_mapping":class_label_mapping, "type":self.loc_type, "initial_values":initial_conds, 
                "location_constants":self.location_constants}
    

class Locations:
    def __init__(self, defined_classes, distance_function):
        self.coords =[[],[]]
        self.locations = []
        self.defined_classes = defined_classes
        self.distance_function = distance_function

        self.type_register = {}
        
    def checkLocationClassesDefined(self):
        for location in self.locations:
            for location_class_label in location.class_labels:
                if location_class_label not in self.defined_classes:
                    raise ValueError(f"Class {location_class_label} not defined (Location name: {location.name})")
        return True
    
    def returnAllLocationConstantNames(self):
        constant_names = set([])
        for location in self.locations:
            constant_names.update(location.returnConstantNames())
        return list(constant_names)
    
    def computeAndSaveDistances(self):
        self.coords =  np.array(self.coords)
        distance_matrix =  self.distance_function(self.coords[0,:], self.coords[1,:])
        for i, x in enumerate(self.locations):
            x.setDistances(list(distance_matrix[i,:].flatten()))

    def returnLocationsDict(self):
        self.checkLocationClassesDefined()
        self.computeAndSaveDistances()

        return {str(i):x.returnDictDescription() for i, x in enumerate(self.locations)}
    
    def addCoordinates(self, lat:float, long:float):
        self.coords[0].append(lat)
        self.coords[1].append(long)
    
    def registerTypeClass(self, type_str:str, type_class):
        if type_str in self.type_register:
            raise(ValueError(f"Type {type_str} already has an associated class {str(self.type_register[type_str])}"))
        else:
            self.type_register[type_str] = type_class

    def addLocation(self, location):
        if isinstance(location, Location):
            if location.loc_type in self.type_register:
                # Want explicit type here rather than inclusive of subclasses
                if type(location) == self.type_register[location.loc_type]:
                    self.locations.append(location)
                    self.addCoordinates(location.lat, location.long)
                else:
                    raise(TypeError(f"Location type {location.loc_type} is already associated with class {self.type_register[location.loc_type]}, not class {type(location)}"))
            else:
                self.registerTypeClass(location.loc_type, type(location))
                self.locations.append(location)
                self.addCoordinates(location.lat, location.long)
        else:
            raise(TypeError(f"location is not a child type of Location base class (type: {type(location)})"))
        
    def addLocations(self, locations):
        for location in locations:
            self.addLocation(location)


# Used for propensity caching - given a rule, find all rules that require an updated propensity

# Use stoichiometry information to determine which
import sympy
import collections

def classToRuleDict(rules, locations, matched_indices, base_classes):
    ctr_dict = collections.defaultdict(set)
    base_ctr_dict = collections.defaultdict(set)
    for rule_i in range(len(matched_indices)):
        rule = rules[rule_i]
        for index_set_i in range(len(matched_indices[rule_i])):
            for slot_i, loc_i in enumerate(matched_indices[rule_i][index_set_i]):
                loc_classes_num = len(locations[loc_i].class_values)
                loc_symbols = rule.sympy_formula[slot_i].atoms(sympy.Symbol)
                #rtc_dict[f"{rule_i} {index_set_i}"].update([f"{str(symbol)} {loc_i}" for symbol in loc_symbols])
                for symbol in loc_symbols:
                    class_index = int(str(symbol)[1:])
                    if class_index < loc_classes_num:
                        ctr_dict[f"{str(symbol)} {loc_i}"].update([f"{rule_i} {index_set_i}"])
                    else:
                        base_ctr_dict[base_classes[loc_classes_num-class_index]].update([f"{rule_i} {index_set_i}"])
    return ctr_dict, base_ctr_dict

# Use propensity infomation to determine which rules require updating based on a change in class value
def ruleToClassesDict(rules, matched_indices):
    rtc_dict = collections.defaultdict(set)
    for rule_i in range(len(matched_indices)):
        rule = rules[rule_i]
        for index_set_i in range(len(matched_indices[rule_i])):
            for slot_i, loc_i in enumerate(matched_indices[rule_i][index_set_i]):
                #rtc_dict[f"{rule_i} {index_set_i}"].update([f"{str(symbol)} {loc_i}" for symbol in loc_symbols])
                rtc_dict[f"{rule_i} {index_set_i}"].update([f"x{i} {loc_i}" for i in range(len(rule.stoichiometry[slot_i])) if not rule.stoichiometry[slot_i][i] == 0])
    return rtc_dict

# Returns a dictionary that maps a rule index to a set of all rule indices that have a changed propensity after a rule trigger.
def ruleToRule(rtc_dict, ctr_dict):
    rtr_dict = {rule_location_key : set([])for rule_location_key in list(rtc_dict.keys())}

    for rule_location_key in list(rtr_dict.keys()):
        for class_loc in rtc_dict[rule_location_key]:
            rule_index = [index.split(" ") for index in ctr_dict[class_loc]]
            rtr_dict[rule_location_key].update([(int(index[0]),int(index[1])) for index in rule_index])
    return rtr_dict

def returnOneStepRuleUpdates(rules, locations, matched_indices, base_classes):
    rtc_dict = ruleToClassesDict(rules, matched_indices)
    ctr_dict, base_ctr_dict = classToRuleDict(rules, locations, matched_indices, base_classes)
    return ruleToRule(rtc_dict, ctr_dict)|base_ctr_dict



class Classes:
    def __init__(self, allow_base_model_classes = True):
        self.class_def_dict = {}
        self.class_names = set([])
        self.base_model_string = "model_"
        if allow_base_model_classes:
            for built_in_class in self.returnBuiltInClasses():
                self.addClass(*built_in_class)
    
    def addClass(self, class_name, class_measurement_unit, class_restriction = "None"):
        if class_name in self.base_model_string:
            raise(ValueError(f"Class: {class_name} cannot contain model prefix{self.base_model_string}"))
        elif not class_name in self.class_names:
            self.class_def_dict[class_name] = {"class_measurement_unit":class_measurement_unit, "class_restriction":class_restriction}
        else:
            raise(ValueError(f"Duplicate type {class_name} provided."))
        
    def returnBuiltInClasses(self):
        builtin_classes = [[f"{self.base_model_string}day", "Day of year"], 
                           [f"{self.base_model_string}hour","Hour of day", "integer"], 
                           [f"{self.base_model_string}months", "Month of year", "integer"]]
        months = [["jan", "January"], ["feb", "February"], ["mar", "March"], ["apr", "April"],
                  ["may", "May"], ["jun", "June"], ["jul", "July"], ["aug", "August"], 
                  ["sept", "September"], ["oct", "October"], ["nov", "November"], ["dec", "December"]
                  ]
        for month in months:
            builtin_classes.append([f"{self.base_model_string}month_{month[0]}", month[1], "indictator"])
        builtin_classes.sort()
        return builtin_classes
    
    def returnClassDict(self):
        return self.class_def_dict


import datetime

import pyRBM.Core.Model as Model
import pyRBM.Build.Locations as ModelLocations
import pyRBM.Build.RuleTemplates as BasicRules
import pyRBM.Simulation.Solvers as Solvers
# seeding_rate in tonnes/ha (hectacre is 2.47105 acres)
# 39.3679 bushels of wheat is a tonne 
model_constants = {
    "Wheat_seeding_rate":0.11,
    "Cereals_seeding_rate":0.2,
    "Barley_seeding_rate":0.3,
    "Potatoes_seeding_rate":0.1,
    "Wheat_germination_rate":1
}

supplyChainClasses = [["NH4", "tonnes"], ["N2", "m^3"], ["H2", "m^3"], ["CH4", "m^3"]]

class FarmRegion(ModelLocations.Location):
    def __init__(self, crop_list:list, lat:float, long:float, name:str, constants = None):
        # Sets lat/long and creates and empty set of compartment labels.
        super().__init__(lat, long, name, loc_type="FarmRegion", constants=constants)
        # Crops exist in three stages in this simplified model: planted, growing and harvested.
        for crop in crop_list:
            self.class_labels.add(crop[0])

# Crop classes:
# Biggest crops as example, from defra https://www.gov.uk/government/statistics/agriculture-in-the-united-kingdom-2022/chapter-7-crops
crops = ["Wheat", "Cereals", "Barley", "Potatoes"]

for crop in crops:
    supplyChainClasses +=  [[f"Seeds_{crop}", "Tonnes"], [f"Planted_{crop}", "Hectacres"], [f"Growing_{crop}", "Hectacres"], [f"Viable_{crop}", "Tonnes"], [f"Harvested_{crop}", "Tonnes"]]

def returnCropRules():    
    rules = []

    # Crop rules
    # TODO add seasonallity.
    # TODO limit based on supply.
    for crop in crops:
        # SIMPLIFIED RULE FOR BASE MODEL
        # Fix requirement of non None propensity class.
        purchase_seeds = BasicRules.ExitEntranceRule(target="FarmRegion", transport_class=f"Seeds_{crop}", transport_amount=1, propensity="0.1", propensity_classes=[f"Seeds_{crop}"],
                                                           rule_name="Purchase Seeds")

        sow_crop = BasicRules.SingleLocationProductionRule(target="FarmRegion",
                                                                        reactant_classes=[f"Seeds_{crop}"], reactant_amount=[1], 
                                                                        product_classes=[f"Planted_{crop}"], product_amount=[model_constants[f"{crop}_seeding_rate"]], propensity=f"Seeds_{crop}",
                                                                        propensity_classes=[f"Seeds_{crop}"], rule_name=f"Sow {crop}")
        
        germinating_crop = BasicRules.SingleLocationProductionRule(target="FarmRegion",
                                                                        reactant_classes=[f"Planted_{crop}"], reactant_amount=[1], 
                                                                        product_classes=[f"Growing_{crop}"], product_amount=[1], propensity=f"0.1*Planted_{crop}",
                                                                        propensity_classes=[f"Planted_{crop}"], rule_name=f"Germinate {crop}")
        # TODO make dependant on Nitrogen levels ect ect.
        crop_growth = BasicRules.SingleLocationProductionRule(target="FarmRegion",
                                                                        reactant_classes=[f"Growing_{crop}"], reactant_amount=[1], 
                                                                        product_classes=[f"Viable_{crop}"], product_amount=[1], propensity=f"Growing_{crop}",
                                                                        propensity_classes=[f"Growing_{crop}"], rule_name=f"Germinate {crop}")
        
        # May need to make propensity based on planted field size, will need to check.
        harvest_crop = BasicRules.SingleLocationProductionRule(target="FarmRegion",
                                                                        reactant_classes=[f"Viable_{crop}"], reactant_amount=[1], 
                                                                        product_classes=[f"Harvested_{crop}"], product_amount=[1], propensity=f"Viable_{crop}",
                                                                        propensity_classes=[f"Viable_{crop}"], rule_name=f"Harvest {crop}")
        rules += [purchase_seeds, sow_crop, germinating_crop, crop_growth, harvest_crop]

    return rules
def supplyChainLocations():
    # Billingham terminal - Produces Ammonium nitrate from Ammonia
    # https://www.cfindustries.com/newsroom/2023/billingham-ammonia-plant
    # https://www.cfindustries.com/what-we-do/fertilizer
    all_locations = []
    # Midpoints from wikipedia, South East and London uses South East midpoint (combined to match DEFRA reporting)
    region_infos = [[54.075, -2.75, "North East"], [55, -1.87, "North West"], [53.566667, -1.2, "Yorkshire & The Humber"], [52.98, -0.75, "East Midlands"], [52.478861, -2.256306, "West Midlands"], 
                    [52.24, 0.41, "East of England"], [51.3, -0.8, "South East & London"], [50.96, -3.22, "South West"], [56.816738, -4.183963, "Scotland"], [52.33022, -3.766409,"Wales"]]
    for region_info in region_infos:
        all_locations.append(FarmRegion(supplyChainClasses, *region_info))

    
    return all_locations

model = Model.Model("Basic Crop")
model.buildModel(supplyChainClasses, supplyChainLocations, returnCropRules, write_to_file = True, save_model_folder="Tests/ModelFiles/")

model.initializeSolver(Solvers.GillespieSolver)
start_date = datetime.datetime(2001, 8, 1)
model.simulate(start_date, 100)

model.trajectory.plotAllClassesOverTime(1)


import numpy as np
import sympy

def returnSympyClassVarsDict(classes):
    symbols_str = ""
    for class_label in classes:
        symbols_str += class_label+" "
    symbols = sympy.symbols(symbols_str)
    if not isinstance(symbols, (list, tuple)):
        symbols = [symbols]
    symbols_dict = {class_label:symbols[index] for index, class_label in enumerate(classes)}
    return symbols_dict


class Rule:
    def __init__(self, rule_name:str, targets:list):

        self.rule_name = rule_name

        self.targets = targets
        # Stochiometries/propensities are with respect
        self.stoichiometies = {i:None for i in range(len(targets))}
        self.propensities = {i:None for i in range(len(targets))}

        self.rule_classes = {i:None for i in range(len(targets))}
        self.stoichiometry_classes = {i:None for i in range(len(targets))}
        self.propensity_classes = {i:None for i in range(len(targets))}

    def addLinearStoichiomety(self, target_indices, stoichiometies, required_target_classes):
        assert(len(target_indices) == len(stoichiometies))
        for i, index in enumerate(target_indices):
            stoichiometry = stoichiometies[i]
            if not self.stoichiometies[index] is None:
                raise(ValueError(f"Overwriting already set stoichiomety is forbidden. Target location {self.targets[index]} at position {str(index+1)}"))
            elif isinstance(stoichiometry, (np.ndarray, list)):
                self.stoichiometies[index] = list(stoichiometry)
                self.stoichiometry_classes[index] = required_target_classes[i]
            else:
                raise(ValueError(f"Unrecognised stoichiomety of type {type(stoichiometies[i])}, for target index {index}"))
    
    def addSimplePropensityFunction(self, target_indices, values, required_target_classes):
        # Accepts matrix or constant values at the moment
        assert(len(target_indices) == len(values))
        for i, index in enumerate(target_indices):
            value = values[i]
            print(f"{value}")
            if not self.propensities[index] is None:
                raise(ValueError(f"Overwriting already set propensity is forbidden. Target location {self.targets[index]} at position {str(index+1)}"))

            if not isinstance(value, str):
                raise(ValueError(f"Unrecognised propensity function of type {type(values[i])}, for target index {index}"))

            self.propensity_classes[index] = required_target_classes[i]
            self.propensities[index] = value

    def validateFormula(self, formula, class_symbols, safe_num = 1):
        # Evaluate when all classes are 0
        subsitution_dict = {}
        for class_str in list(class_symbols.keys()):
            subsitution_dict[class_symbols[class_str]] = safe_num
        res = formula.evalf(subs=subsitution_dict)
        # We require that a numerical result is outputted and no symbols are left over.
        print(res)
        assert(isinstance(res, (sympy.core.numbers.Float, sympy.core.numbers.Zero)))
        return True

    def checkRuleDefinition(self, builtin_class_symbols, locations_constant_symbols):
        for i in range(len(self.targets)):
            if self.stoichiometies[i] is None:
                raise(ValueError(f"Incorrect rule definition for {self.rule_name}\nThe Location type {self.targets[i]} at rule position {str(i+1)} has no defined stochiometry."))
            elif self.propensities[i] is None:
                raise(ValueError(f"Incorrect rule definition for {self.rule_name}\nThe Location type {self.targets[i]} at rule position {str(i+1)} has no defined propensity."))
            elif self.propensity_classes[i] is None:
                raise(ValueError(f"Incorrect rule definition for {self.rule_name}\nThe Location type {self.targets[i]} at rule position {str(i+1)} has no defined propensity class requirement."))
            elif self.stoichiometry_classes[i] is None:
                raise(ValueError(f"Incorrect rule definition for {self.rule_name}\nThe Location type {self.targets[i]} at rule position {str(i+1)} has no defined stochiometry class requirement."))
            
        for index in range(len(self.propensities)):
            symbols = returnSympyClassVarsDict(self.propensity_classes[index]) | builtin_class_symbols | locations_constant_symbols
            sympy_formula = sympy.parse_expr(self.propensities[index], local_dict=symbols)
            self.validateFormula(sympy_formula, symbols)
            
    def mergeClassLists(self):
        # Maps new index to class label
        self.rule_classes = []
        for i, target in enumerate(self.targets):
            sorted_classes = sorted(set(self.propensity_classes[i] + self.stoichiometry_classes[i]))
            tmp_rule_class_dict = {i:comp_class for i, comp_class in enumerate(sorted_classes)}
            # Might be better to remap inputs for more complex functions rather than directly changing the definition of the function.
                # Use additive identity here
            new_stoichiometry = np.zeros(len(sorted_classes))
            for h, old_class in enumerate(self.stoichiometry_classes[i]):
                for j in range(len(tmp_rule_class_dict)):
                    if old_class == tmp_rule_class_dict[j]:
                        new_stoichiometry[j] = self.stoichiometies[i][h]
            self.stoichiometies[i] = list(new_stoichiometry)
            self.rule_classes.append(tmp_rule_class_dict)
        
    def returnRuleDict(self):
        # Check performed in Rules class
        # self.checkRuleDefinition(builtin_class_symbols, locations_constants)
        self.mergeClassLists()

        # UNPACK DEF TO SPECIFIC LOCATION
        rule_dict = {"name":self.rule_name, "target_types":self.targets, "required_classes":self.rule_classes,
                     "stoichiometries":self.stoichiometies, "propensities":self.propensities}
        return rule_dict

class Rules:
    def __init__(self, defined_classes, location_constants : list):
        self.rules = []
        self.defined_classes = defined_classes
        self.model_prefix = "model_"
        self.location_prefix = "loc_"

        builtin_classes = []

        for classes in defined_classes:
            if self.model_prefix in classes:
                builtin_classes.append(classes)


        self.builtin_symbols = returnSympyClassVarsDict(builtin_classes)
        # DOESN'T CHECK FOR EXISTENCE OF CONSTANTS FOR EACH LOCATION - THIS IS DONE IN THE RULE MATCHING
        self.location_constants_symbols = returnSympyClassVarsDict(location_constants)
    
    def addRule(self, rule:Rule):
        if isinstance(rule, Rule):
            self.rules.append(rule)
        else:
            raise(TypeError(f"rule is not a child type of Rule base class (type: {type(rule)})"))
        
    def addRules(self, rules):
        for rule in rules:
            self.addRule(rule)
    
    def checkRules(self):
        for rule in self.rules:
            rule.checkRuleDefinition(self.builtin_symbols, self.location_constants_symbols)

            for propensity_class_index in range(len(rule.propensity_classes)):
                location_propensity_class = rule.propensity_classes[propensity_class_index]
                for loc_prop_class in location_propensity_class:
                    if not loc_prop_class in self.defined_classes:
                        raise ValueError(f"Class required for the propensity, {loc_prop_class}, not defined (Rule name: {rule.rule_name})")
                    
            for stoichiometry_class_index in range(len(rule.stoichiometry_classes)):
                location_stoichiometry_class = rule.stoichiometry_classes[stoichiometry_class_index]
                print(location_stoichiometry_class)
                for loc_stoich_class in location_stoichiometry_class:
                    if not loc_stoich_class in self.defined_classes:
                        raise ValueError(f"Class required for the stoichiometry, {loc_stoich_class} not defined (Rule name: {rule.rule_name})")
        return True

    def returnMetaRuleDict(self):
        self.checkRules()
        return {str(i):rule.returnRuleDict() for i, rule in enumerate(self.rules)}


import json
import os

import numpy as np
import pyRBM.Simulation.Rule as Rule
import pyRBM.Simulation.Location as Location

def writeDictToJSON(dict_to_write:dict, filename:str):
    
    folder =''.join([folder+'/' for folder in filename.split("/")[:-1]])
    dir_to_create = os.path.join(os.curdir,folder)

    print(dir_to_create)
    if not os.path.exists(dir_to_create):
        print("no")
        os.makedirs(dir_to_create)

    json_file = json.dumps(dict_to_write, indent=4, sort_keys=True)
    with open(f"{filename}.json", "w+", encoding='utf-8') as outfile:
            outfile.write(json_file)

def readDictFromJSON(filename:str):
    file_data = None
    with open(filename, encoding='utf-8') as infile:
        file_data = json.load(infile)
    return file_data

def processFilenameOrDict(filename:str, provided_dict:dict):
    model_data = None
    if (filename is None) and (provided_dict is None):
        raise(ValueError("Provide either a "))
    elif not filename is None:
        model_data = readDictFromJSON(filename)
    else:
        model_data = provided_dict
    return model_data

def loadLocations(locations_filename:str = None, build_locations_dict:dict = None):
    """ Loads all locations from a model location json (see ModelCreation for details).

    Parameters: 
        - locations_filename: the string of the file location containing the location definitions.
        
    Returns: a list of Locations corresponding to all locations in the location_file
    """
    location_list = []
    locations_data = processFilenameOrDict(locations_filename, build_locations_dict)

    for loc_index in range(len(locations_data)):
        location_dict = locations_data[str(loc_index)]
        location = Location.Location(index=loc_index, name=location_dict["location_name"], lat=location_dict["lat"], long=location_dict["long"], loc_type=location_dict["type"],
                                         label_mapping=location_dict["label_mapping"],
                                         initial_class_values=np.array(location_dict["initial_values"]), location_constants=location_dict["location_constants"])
        location_list.append(location)
    return location_list
    

def loadMatchedRules(locations,  num_builtin_classes, matched_rules_filename:str = None, matched_rule_dict:dict=None):
    """ Loads all rules from a model matched rules json (see ModelCreation for details).

    Parameters: 
        - matched_rules_filename: the string of the file location containing the rule definitions.
    Returns: [a list of rules remapped to all possible location sets, 
              a 2d list of lists of satisfying indices for the corresponding rule]
    """
    rules_list = []
    applicable_indices = []
    rules_data =  processFilenameOrDict(matched_rules_filename, matched_rule_dict)
    
    for rule_index in range(len(rules_data)):
        rules_dict = rules_data[str(rule_index)]
        stochiometries = []
        propensities = []
        # Convert stochiometries and propensities to numpy arrays - need to use a list of arrays as the 2nd dimension of the array has varying dimension.
        for loc_stoichiometry in rules_dict["stoichiomety"]:
            stochiometries.append(np.array(loc_stoichiometry))
        for loc_propensity in rules_dict["propensity"]:
            propensities.append(loc_propensity)

        rule = Rule.Rule(propensity=propensities, stoichiometry=stochiometries, rule_name=rules_dict["rule_name"], num_builtin_classes=num_builtin_classes, locations=locations, rule_index_sets=rules_dict["matching_indices"])
        applicable_indices.append(rules_dict["matching_indices"])
        rules_list.append(rule)
    return [rules_list, applicable_indices]

def loadClasses(model_prefix = "model_", classes_filename:str = None, classes_dict:dict = None):
    class_dict = {}
    built_in_class_dict = {}
    class_data = processFilenameOrDict(classes_filename, classes_dict)

    for class_key in list(class_data.keys()):
        if model_prefix in class_key:
            built_in_class_dict[class_key] = class_data[class_key]
        else:
            class_dict[class_key] = class_data[class_key]
    return [class_dict, built_in_class_dict]




import numpy as np
import sympy

class Rule:
    def __init__(self, propensity, stoichiometry, rule_name, num_builtin_classes, locations, rule_index_sets) -> None:
        assert (len(stoichiometry) == len(propensity))
        if isinstance(propensity, (list)):
            self.lambda_propensities = []
            self.contains_location_constant = []
            self.sympy_formula = []

            for slot_i, formula_str in enumerate(propensity):
                symbol_string = ''.join([f'x{str(i)} ' for i in range(len(stoichiometry[slot_i])+num_builtin_classes)])
                formula_symbols = sympy.symbols(symbol_string, real=True)

                #symbols = sympy.symbols("".join([f"x{i} " for i in range(len(stoichiometry[loc_i]))]) , real=True)
                #M = sympy.IndexedBase('x', shape=(dim))
            
                #symbols = [f"x{i}" for i in range(len(stoichiometry[loc_i]))]

                if not "loc_" in formula_str:
                    formula = sympy.parse_expr(formula_str)
                    self.sympy_formula.append(formula)
                    f=sympy.lambdify(formula_symbols, formula.simplify(), "numpy")
                    self.contains_location_constant.append(False)
                else:
                    applicable_indices = self.findIndices(rule_index_sets, slot_i)
                    #formula_without_constants = self.subsituteConstants(formula_str, {key:"" for key in list(locations[applicable_indices[0]].location_constants.keys())})

                    f = {loc_index: sympy.lambdify(formula_symbols, sympy.parse_expr(self.subsituteConstants(formula_str, locations[loc_index].location_constants)).simplify(), "numpy") for loc_index in applicable_indices}

                    self.sympy_formula.append(sympy.parse_expr(self.subsituteConstants(formula_str, locations[applicable_indices[0]].location_constants)))

                    self.contains_location_constant.append(True)

                self.lambda_propensities.append(f)
            self.propensity_function = lambda x, loc : np.dot(x, self.propensity_matrix[loc])
        else:
            raise(ValueError("Unsupported Propensity in Model Loading"))
        self.rule_name = rule_name
        self.stoichiometry = stoichiometry
        self.contains_location_constant = np.array(self.contains_location_constant)
    
    def subsituteConstants(self, formula_str:str, location_constants:dict):
        out_formula = formula_str
        for loc_constant in list(location_constants.keys()):
            out_formula = out_formula.replace(loc_constant, str(location_constants[loc_constant]))
        return out_formula

    def findIndices(self, rule_index_sets, slot_index):
        possible_indices = set([])

        for index_set in rule_index_sets:
            possible_indices.add(index_set[slot_index])

        return list(possible_indices)
    def locationAttemptedCompartmentChange(self, class_values, location_index, times_triggered):
        new_values = class_values + times_triggered*self.stoichiometry[location_index]

        return new_values
    
    def returnPropensity(self, locations, builtin_classes):
        assert(len(locations) == len(self.stoichiometry))
        # Assume product operation.
        propensity = 1
        for loc_i, location in enumerate(locations):
            # Apply thresholding here to ensure that no negative propensities are used.
            # print(self.lambda_propensities[loc_i](*location.class_values, *builtin_classes))
            if not self.contains_location_constant[loc_i]:
                propensity *= max(0, self.lambda_propensities[loc_i](*location.class_values, *builtin_classes))
            else:
                propensity *= max(0, self.lambda_propensities[loc_i][location.index](*location.class_values, *builtin_classes))

        assert (propensity >= 0)
        return propensity
    
    # We expect pure Gillespie to have 0 propensity for negative rule changes, however with Tau leaping we may need
    # to check whether a series of rule changes lead 
    def triggerAttemptedRuleChange(self, locations, times_triggered = 1):
        assert(len(locations) == len(self.stoichiometry))
        negative = False
        new_class_values = []

        for loc_i, location in enumerate(locations):
            new_location_values = self.locationAttemptedCompartmentChange(location.class_values, loc_i, times_triggered)
            # CHANGE TODO
            if np.any(new_location_values<-1):
                negative = True
                break
            else:
                new_class_values.append(new_location_values)
        if not negative:
            for loc_i, location in enumerate(locations):
                location.updateCompartmentValues(new_class_values[loc_i])
        return not negative
        


""" Generic framework to match metarules to locations, find all location indices that match and rewrite the propensities and stoichiometries to use array indices.
"""

import numpy as np

def isSubtypeOf(parent_type, child_type):
    """ Indicates whether the "child_type" rule type is a subtype of the parent_type.
    
    Current returns parent_type == child_type.

    In the future this should be determined by a tree based hierachical graph.
    """
    # Equality for the moment
    return parent_type == child_type

def returnRuleMatchingIndices(rules, locations):
    # For each rule, on each type that matches a general type
    filled_rules = {str(i):[] for i in range(len(rules))}
    for rule_i in range(len(rules)):
        rule = rules[str(rule_i)]
        matchedTypeToIndices = {i:[] for i in range(len(rule["target_types"]))}
        # Check all locations to see which locations correspond to which required types by the rule.
        # (note, a rule may correspond to multiple required types but it can only be used in one slot).

        # We obtain a dictionary mapping each required location type to a list of fufilling indices in our location set.
        for location_i in range(len(locations)):
            for rule_targets_i in range(len(rule["target_types"])):
                if isSubtypeOf(rule["target_types"][rule_targets_i], locations[str(location_i)]["type"]):
                    matchedTypeToIndices[rule_targets_i].append(location_i)
        # From the previously described location set, obtain a tuple of indices (if it exists) that satisfies the rule.
        # Do this iteratively, loop over all index sets in construction and add the new index if it isn't already inside.
        print(f"MATCHED {matchedTypeToIndices}")
        rule_indices = {}
        for rule_targets_i in range(len(rule["target_types"])):
            atleast_one_satisifying = False
            if len(rule_indices) == 0:
                if len(matchedTypeToIndices[0])>0:
                    for mtti in matchedTypeToIndices[0]:
                        location = locations[str(location_i)]
                        if not location["type"] in  rule_indices.keys():
                             rule_indices[location["type"]] = []
                        rule_indices[location["type"]].append([mtti])
                    atleast_one_satisifying = True
            else:
                # Every index that can be subbed in at that place.
                # We keep a dictionary here so we can map explict types.
                correct_length_rules = {}
                for loc_index in matchedTypeToIndices[rule_targets_i]:
                    rule_keys = list(rule_indices.keys())
                    # Which explict types are being used.
                    for rule_key in rule_keys:
                        for index_matching in rule_indices[rule_key]:
                            if not loc_index in index_matching:
                                if not rule_key+"_"+locations[loc_index]["type"] in list(correct_length_rules.keys()):
                                    correct_length_rules[rule_key+"_"+locations[loc_index]["type"]] = [index_matching+[loc_index]]
                                else:
                                    correct_length_rules[rule_key+"_"+locations[loc_index]["type"]].append(index_matching+[loc_index])
                                atleast_one_satisifying = True
                rule_indices = correct_length_rules

            # Prune rules that will never able to be completed to reduce size.
            #correct_length_rules = [] 
            #for rule_index in rule_indices:
                #if len(rule_index) == rule_targets_i+1:
                    #correct_length_rules.append(rule_index)

            if not atleast_one_satisifying:
                raise(ValueError(f"Rule {rule_i} has no satisying location for required type index{rule_targets_i}, type {str(rule['target_types'][rule_targets_i])}. Rule will never be trigger - remove rule"))
        filled_rules[str(rule_i)] = rule_indices
    return filled_rules

# Return the final propensity for a given rule provided concrete locations. 
# Assumption that locations of the same type have the same compartments.
def obtainPropensity(rule, locations, builtin_classes):
    propensities = rule["propensities"]
    new_propensities = []
    for location_i, location in enumerate(locations):
        new_propensity = propensities[location_i]
        new_label_mapping = location["label_mapping"]
        # Order: model var, location const, location class
        for built_in_i, builtin_class in enumerate(builtin_classes):
            new_propensity = new_propensity.replace(builtin_class[0], f"x{built_in_i+len(new_label_mapping)}")
            

        for label_i in list(new_label_mapping.keys()):
            new_propensity = new_propensity.replace(new_label_mapping[label_i], f"x{label_i}")
        # Add the built in class at the end of the location classes

        new_propensities.append(new_propensity)
    print(new_propensities)
    return new_propensities

def obtainStochiometry(rule, locations):
    """ Remaps the rule to fit the class size found in each location
    
    Example: Original Rule mapping {0:"Class1", 1:"Class2"} (size of input to Stochiometry and Propensity functions: 2)
    Location and Fitted Rule mapping {0:"ClassA", 1:"Class1", 2:"ClassB", 3:"Class2"}  (size of input to Stochiometry and Propensity functions: 4)


    Parameters:
        - rule: the Rule that we are remapping the propensity and stochiometry to fit each location.
        - locations: location instances (a list the length of rule.types) that fit each of the rule.types.

    Returns: list of expanded stochiometries for all rule locations  
    """
    # Rule
    stoichiometries = rule["stoichiometries"]

    current_class_mapping = rule["required_classes"]

    new_stoichiometries = []
    # Locations
    for rule_location, location in enumerate(locations):
        new_label_mapping = location["label_mapping"]
        new_stoichiometry = np.zeros(len(new_label_mapping))

        for rule_class_index in range(len(current_class_mapping[rule_location])):
            class_found = False
            for location_class_index in range(len(new_label_mapping)):
                if current_class_mapping[rule_location][rule_class_index] == new_label_mapping[str(location_class_index)]:
                    new_stoichiometry[location_class_index] = stoichiometries[rule_location][rule_class_index]
                    # Only needs to happen once so can be placed here, possibly refactor
                    class_found = True
            if not class_found:
                raise(ValueError("Rule class not found in location class in rule matching (missing stoichiometry class)."))
        new_stoichiometries.append(list(new_stoichiometry))
        
    return new_stoichiometries

def returnMatchedRulesDict(rules, locations, builtin_classes):
    matched_rules = returnRuleMatchingIndices(rules, locations)
    builtin_classes = sorted(builtin_classes)
    concrete_match_rules_dict = {}
    concrete_rules = 0
    for rule_i in range(len(matched_rules)):
        rule = rules[str(rule_i)]
        matched_rule = matched_rules[str(rule_i)]
        concrete_rule_types = list(matched_rule.keys())
        for concrete_rule_type in concrete_rule_types:
            concrete_rule_dict = {"rule_num":rule_i, "rule_name":rule["name"],
                                  "rule_location_types":concrete_rule_type, "matching_indices":matched_rule[concrete_rule_type]}
            # Assume all locations have the same classes - will be asserted in later versions.
            example_locations = []
            # Take the first set as an example
            for location_index in matched_rule[concrete_rule_type][0]:
                example_locations.append(locations[str(location_index)])
            
            # TODO ensure compatibility with further propensity functions
            concrete_rule_dict["stoichiomety"] = obtainStochiometry(rule, example_locations)
            concrete_rule_dict["propensity"] = obtainPropensity(rule, example_locations, builtin_classes)
            concrete_match_rules_dict[str(concrete_rules)] = concrete_rule_dict
            concrete_rules+= 1

    return concrete_match_rules_dict


import Application
import datetime

start_date = datetime.datetime(2001, 8, 1)

epi_application = Application.ModelBackend(start_date=start_date, model_folder="Backend/ModelFiles/CropModelv1/", propensity_caching=True)
out = epi_application.simulate(time_limit=365, max_iterations=1000000)

out.plotAllClassesOverTime(location_index=1)


import numpy as np
from math import radians, cos, sin, asin, sqrt

# Credit to https://stackoverflow.com/questions/29545704/fast-haversine-approximation-python-pandas

def haversine(lon1, lat1, lon2, lat2):
    """
    Calculate the great circle distance between two points 
    on the earth (specified in decimal degrees)
    """
    # convert decimal degrees to radians
    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])
    # haversine formula
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat/2)**2 + cos(lat1) * cos(lat2) * sin(dlon/2)**2
    c = 2 * asin(sqrt(a))
    km = 6367 * c
    return km

def createEuclideanDistanceMatrix(lats, longs):
    # Only need to find the upper right triangle.
    distm = np.zeros((len(lats), len(lats)))
    for i in range(len(lats)):
        for h in range(i, len(longs)):
            distm[i][h] = haversine(longs[i], lats[i], longs[h],  lats[h])
    # Fill in the lower left triangle
    distm = distm.T+distm
    return distm



import pandas as pd
import pyRBM.Build.Utils as Utils
import pyRBM.Build.RuleTemplates as RuleTemplates
from collections import defaultdict

#def process

def extractColumns(example_row, exlude_list:list = []):
    cols = []
    all_keys = example_row.keys()
    for col_i in range(len(all_keys)):
        col = all_keys[col_i]
        if not col in exlude_list:
            cols.append(col)
    return cols

def createMonthPropTerm(months:list):
    term_str = "("

    month_prefix = "model_month_"
    for month in months:
        if not term_str == "(":
            term_str += "+"
        term_str += f"{month_prefix}{month}"
    term_str += ")"

    return term_str

class CropStages:
    def __init__(self, stages_filepath:str, crops_filepath:str) -> None:
        crop_stages_data = pd.read_csv(stages_filepath, converters={})
        crops_background_data = pd.read_csv(crops_filepath)
        

        self.location_type_name = "FarmRegion"
        self.crop_stages_dict = defaultdict(list)
        self.crops = set([])
        self.crop_stage_info = {}
        self.additional_classes = ["Seeds", "Planted", "Harvested"]

        self.crop_info = {}

        self.stage_cols = []
        for i in range(len(crop_stages_data)):
            stage_row = crop_stages_data.iloc[i]
            if len(self.stage_cols) == 0:
                self.stage_cols = extractColumns(stage_row, exlude_list=["Crop","Growth Stage"])
            crop_name = stage_row["Crop"]
            crop_stage = stage_row["Growth Stage"]

            self.crops.add(crop_name)
            self.crop_stages_dict[crop_name].append(crop_stage)

            self.crop_stage_info[f"{crop_name}_{crop_stage}"] = {cols:stage_row[cols] for cols in self.stage_cols}
        
        for i in range(len(crops_background_data)):
            crop_row = crops_background_data.iloc[i]
            crop = crop_row["Crop"]
            self.crop_cols = []
            if len(self.crop_cols) == 0:
                self.crop_cols = extractColumns(crop_row, exlude_list=["Crop"])
            self.crop_info[f"{crop}"] = {cols:crop_row[cols] for cols in self.crop_cols}


        for crop in self.crops:
            self.crop_stages_dict[crop] += self.additional_classes
            for stage in self.additional_classes:
                self.crop_stage_info[f"{crop}_{stage}"] = {"Measurement Units":"Placeholder"}

    def returnBasicStageDict(self):
        return self.crop_stages_dict

    def returnCropRules(self):
        # Seeds, harvested, yield
        rules = []

        for crop in self.crops:
            last_crop_stage = None
            crop_info_dict = self.crop_info[crop]
            stages = self.crop_stages_dict[crop]
            for stage in stages:
                if not stage in self.additional_classes:
                    current_class_string = f"{crop}_{stage}"
                    
                    info = self.crop_stage_info[current_class_string]
                    target_str = None
                    if info["Next Growth Stage"] == "" or info["Next Growth Stage"] == " ":
                        if last_crop_stage is None:
                            last_crop_stage = stage
                        else:
                            raise(ValueError(f"Crop {crop} has atleast two final crop stages {last_crop_stage} and {stage}"))
                        target_str = crop+"_"+"Harvested"
                    else:
                        target_str = crop+"_"+info["Next Growth Stage"]

                    crop_growth_rule  = RuleTemplates.SingleLocationProductionRule(self.location_type_name,
                                                                        [crop+"_"+stage],[info["Transport Base Amt"]/info["Stochasticity"]],
                                                                        [target_str],[info["Transport Base Amt"]*info["Next Stage Measurement Factor"]/info["Stochasticity"]],
                                                                        f"{current_class_string}*{info['Stochasticity']}", [current_class_string],
                                                                        f"{crop} {stage} to {info['Next Growth Stage']} Crop Growth")
                    crop_death_rule = RuleTemplates.ExitEntranceRule(self.location_type_name,
                                                                        crop+"_"+stage,info["Transport Base Amt"],
                                                                        f"{current_class_string}*{info['Spoilage Rate']}", [current_class_string],
                                                                        f"{crop} {stage} Decay")
        
                    rules.append(crop_growth_rule)
                    rules.append(crop_death_rule)

            seed_period_term = createMonthPropTerm(crop_info_dict["Plant Period"].split(" "))
            first_stage = crop+"_"+crop_info_dict["First Stage"]
            seed_rule = RuleTemplates.SingleLocationProductionRule(self.location_type_name,
                                                                        [crop+"_Seeds"],[1000],
                                                                        [first_stage,crop+"_Planted"],[1000, 1000],
                                                                        f"{seed_period_term}*{crop}_Seeds*(loc_{crop_info_dict['Capacity Constant Name']} - {crop}_Planted)", [f"{crop}_Planted", f"{crop}_Seeds"],
                                                                         f"{crop} planting")
            harvest_period_term = createMonthPropTerm(crop_info_dict["Harvest Period"].split(" "))
            harvest_rule = RuleTemplates.SingleLocationProductionRule(self.location_type_name,
                                                                        [crop+"_"+last_crop_stage],[1],
                                                                        [crop+"_Harvested"],[1],
                                                                        f"{harvest_period_term}*{crop+'_'+last_crop_stage}", [current_class_string],
                                                                        f"Productive {crop} harvest")
            #for harvest_classes = [crop+harvest for harvest in []]
            #harvest_rules = BasicRules.ExitEntranceRule(self.location_type_name,
            #                                                            [crop+"_"+last_crop_stage],[1],
             #                                                           [crop+"_harvested"],[1],
            #                                                            f"{harvest_period_term}*{info['Stochasticity']}", [current_class_string])
            rules.append(seed_rule)
            rules.append(harvest_rule)
        return rules
    
    def returnCropClassData(self):
        crop_classes = [[f"{crop}_{stage}", self.crop_stage_info[f"{crop}_{stage}"]["Measurement Units"],"None"]
                        for crop in list(self.crop_stages_dict.keys()) for stage in self.crop_stages_dict[crop]]
        return crop_classes
class CropRegions:
    def __init__(self, region_filepath:str, crop_stages:CropStages) -> None:
        region_data = pd.read_csv(region_filepath)
        self.regions = []
        constants = []
        for i in range(len(region_data)):
            region_row = region_data.iloc[i]
            if len(constants) == 0:
                constants = extractColumns(region_row, exlude_list=["Name","Lat","Long"])
            
            print(region_row["Lat"])
            print(crop_stages.returnBasicStageDict())
            region = Utils.FarmRegion(crop_stages.returnBasicStageDict(), region_row["Lat"], region_row["Long"], region_row["Name"])
            print(region_row.keys()[0])
            region.addAndSetConstants({constant.replace(" (Ha)","").replace(" ", "_"):float(region_row[constant]) for constant in constants})
            
            region.setInitialConditions({f"{crop}_Seeds":float(crop_stages.crop_info[crop]["Starting Seeds"]) for crop in crop_stages.crops})

            self.regions.append(region)
        
    def returnRegions(self):
        return self.regions


# This file is used to allow

from datetime import timedelta, datetime
import numpy as np

class ModelState:
    def __init__(self, model_classes, start_datetime):
        self.MONTHS =  ["jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sept", "oct", "nov", "dec"]
        self.model_prefix = "model_"

        self.IMPLEMENTED_MODEL_CLASSES = [f"model_{var}" for var in ["day", "hour", "months"]]
        self.IMPLEMENTED_MODEL_CLASSES += [f"model_month_{month}" for month in self.MONTHS]
        self.model_classes = {}

        self.time_measurement = "days"

        for model_class in model_classes:
            if model_class in self.IMPLEMENTED_MODEL_CLASSES:
                self.model_classes[model_class] = None
                print(f"{model_class} implemented")
            else:
                self.model_classes[model_class] = None
                raise(ValueError(f"Model class {model_class} not implemented"))


        self.elapsed_time = 0
        self.iterations = 0
        self.current_month = None

        self.changed_vars = [classes for classes in list(self.model_classes.keys())]


        if isinstance(start_datetime, datetime):
            self.start_datetime = start_datetime
            self.current_datetime = start_datetime
        else:
            raise(ValueError(f"Start date must be of type datetime not {type(start_datetime)}"))
        self.initaliseCalendarInfo()
        
    def reset(self):
        self.elapsed_time = 0
        self.iterations = 0
        self.current_datetime = self.start_datetime
        self.changed_vars = [classes for classes in list(self.model_classes.keys())]
        self.updateCalendarInfo()

    # Convert the self.current_datetime into model variables used
    def initaliseCalendarInfo(self):
        # Initialise all indicators to zero and perform the standard update step.
        for key in self.IMPLEMENTED_MODEL_CLASSES:
            self.model_classes[key] = 0
        self.updateCalendarInfo()
    
    
    def changeDate(self, new_date):
        self.start_datetime = new_date
        self.reset()

    def updateCalendarInfo(self):

        # https://docs.python.org/3/library/datetime.html#datetime.datetime.timetuple
        current_datetime_info = self.current_datetime.timetuple()

        month_index = current_datetime_info[1]-1

        if self.current_month is None:
            self.model_classes[f"{self.model_prefix}month_{self.MONTHS[month_index]}"] = 1
            self.model_classes[f"{self.model_prefix}months"] =  current_datetime_info[1]
            self.current_month = self.MONTHS[month_index]

        if not self.current_month == self.MONTHS[month_index]:
            self.changed_vars.append(f"{self.model_prefix}month_{self.current_month}")
            self.changed_vars.append(f"{self.model_prefix}month_{self.MONTHS[month_index]}")
            self.changed_vars.append(f"{self.model_prefix}months")

            self.model_classes[f"{self.model_prefix}month_{self.current_month}"] = 0
            self.model_classes[f"{self.model_prefix}month_{self.MONTHS[month_index]}"] = 1
            self.current_month = self.MONTHS[month_index]
            self.model_classes[f"{self.model_prefix}months"] =  current_datetime_info[1]

            # Variables that need their propensities updating

        if not self.model_classes[f"{self.model_prefix}day"] == current_datetime_info[2]:
            self.model_classes[f"{self.model_prefix}day"] = current_datetime_info[2]
            self.changed_vars.append(f"{self.model_prefix}day")
        
        rounded_hours = round(current_datetime_info[4]/60.0, 1)
        if not self.model_classes[f"{self.model_prefix}hour"] == current_datetime_info[3] + rounded_hours:
            self.model_classes[f"{self.model_prefix}hour"] = current_datetime_info[3] + rounded_hours
            self.changed_vars.append(f"{self.model_prefix}hour")


    def processUpdate(self, new_time):
        self.changed_vars = []
        if new_time == None:
            print("Ending model simulation")
            return
        self.updateTime(new_time)
        self.iterations += 1

    def updateTime(self, new_time):
        time_change = None
        time_diff = new_time - self.elapsed_time
        if self.time_measurement == "hours":
            time_change = timedelta(hours=time_diff)
        elif self.time_measurement == "days":
            time_change = timedelta(days=time_diff)
        else:
            raise(ValueError(f"Unsupported time increment type {self.time_measurement}"))
        
        self.current_datetime = self.current_datetime + time_change
        self.elapsed_time = new_time

        self.updateCalendarInfo()
        
    def returnModelClassesValues(self):
        return self.model_classes.values()
    
    def returnModelClasses(self):
        return list(self.model_classes.values())
                    
    def returnChangedVars(self):
        return self.changed_vars


import pyRBM.Build.Classes as Classes
import pyRBM.Build.Locations as Locations
import pyRBM.Build.Rules as Rules
import pyRBM.Build.RuleMatching as RuleMatching
import pyRBM.Build.Utils as Utils

import pyRBM.Core.Cache as Files

import pyRBM.Simulation.State as State
import pyRBM.Simulation.Solvers as Solvers
import pyRBM.Simulation.RuleChain as RuleChain
import pyRBM.Simulation.Trajectory as Trajectory


import time
import datetime


class Model:
    def __init__(self, model_name):
        self.builtin_classes = True
        self.model_name = model_name
        return

        
    def createLocations(self):
        all_locations = Locations.Locations(self.defined_classes, self.distance_func)
        locations = self.create_locations_func()
        all_locations.addLocations(locations)
        # Distance computation done as part of writeJSON - set as location constants
        locations = all_locations.returnLocationsDict()
        self.location_constants = all_locations.returnAllLocationConstantNames()

        return locations

    def createRules(self):
        # Use np.identity(len()) .... for no change
        all_rules = Rules.Rules(self.defined_classes, self.location_constants)

        rules = self.create_rules_func()
        all_rules.addRules(rules)

        return all_rules.returnMetaRuleDict()

    def defineClasses(self):
        classes = Classes.Classes(self.builtin_classes)
        for class_info in self.classes_defintions:
            classes.addClass(*class_info)
        classes_dict = classes.returnClassDict()
        self.defined_classes = classes_dict.keys()

        return classes_dict

    def matchRules(self, rules):
        additional_classes = []
        if self.builtin_classes:
            additional_classes = Classes.Classes().returnBuiltInClasses()
        return RuleMatching.returnMatchedRulesDict(rules, self.locations_dict, additional_classes)
    
    
    def buildModel(self, classes_defintions, create_locations, create_rules, distance_func = Utils.createEuclideanDistanceMatrix,
                   write_to_file = False,  save_model_folder:str = "/ModelFiles/", location_filename:str = "Locations", matched_rules_filename:str = "LocationMatchedRules",
                   classes_filename:str = "Classes", save_meta_rules = False, metarule_filename:str = "MetaRules"):
        
        self.classes_defintions = classes_defintions
        self.create_locations_func = create_locations
        self.create_rules_func = create_rules
        self.distance_func = distance_func
        self.save_model_folder = save_model_folder

        self.classes_dict = self.defineClasses()
        self.locations_dict = self.createLocations()
        rules = self.createRules()
        self.matched_rules_dict = self.matchRules(rules)

        self.write_to_file = write_to_file


        if self.write_to_file:
            if not save_meta_rules:
                metarule_filename = None
            else:
                Files.writeDictToJSON(rules, f"{self.save_model_folder}{self.model_name}/{metarule_filename}")

            self.save_model_folder,self.location_filename,self.matched_rules_filename,self.classes_filename,self.metarule_filename = [save_model_folder,location_filename,
                                                                                                                                      matched_rules_filename,classes_filename,metarule_filename]
            for model_dict, file_loc in [(self.matched_rules_dict, f"{self.save_model_folder}{self.model_name}/{self.matched_rules_filename}"),
                                         (self.classes_dict, f"{self.save_model_folder}{self.model_name}/{self.classes_filename}"),
                                         (self.locations_dict, f"{self.save_model_folder}{self.model_name}/{self.location_filename}")]:
                Files.writeDictToJSON(model_dict, file_loc)
        else:
            if save_meta_rules:
                print("Warning: meta rules not saved as file saving is false")
            else:
                self.save_model_folder,self.location_filename,self.matched_rules_filename,self.classes_filename,self.metarule_filename = [None,None,None,None,None]

        self.convertToSimulation()

    def convertToSimulation(self):
        self.classes, self.builtin_classes = Files.loadClasses(classes_dict=self.classes_dict)
        self.locations = Files.loadLocations(build_locations_dict=self.locations_dict)
        self.rules, self.matched_indices = Files.loadMatchedRules(self.locations, num_builtin_classes=len(self.builtin_classes),
                                                                  matched_rule_dict=self.matched_rules_dict)


        self.trajectory = Trajectory.Trajectory(self.locations)
        self.model_state = State.ModelState(self.builtin_classes, datetime.datetime.now())

        self.model_initialized = True
        self.solver_initialized = False

    def loadModelFromJSONFiles(self, location_filename:str = "Locations", matched_rules_filename:str = "LocationMatchedRules",
                           classes_filename:str = "Classes", model_folder:str = "Backend/ModelFiles/"):
        
        self.matched_rules_filename = matched_rules_filename
        self.location_filename = location_filename
        self.save_model_folder = model_folder
        self.classes_filename = classes_filename

        self.classes, self.builtin_classes = Files.loadClasses(classes_filename = f"{self.save_model_folder}{self.model_name}/{self.classes_filename}")
        self.locations = Files.loadLocations(locations_filename = f"{self.save_model_folder}{self.model_name}/{self.location_filename}")
        self.rules, self.matched_indices = Files.loadMatchedRules(self.locations, num_builtin_classes=len(self.builtin_classes), 
                                                                  matched_rules_filename= f"{self.save_model_folder}{self.model_name}/{self.matched_rules_filename}")


        self.trajectory = Trajectory.Trajectory(self.locations)
        self.model_state = State.ModelState(self.builtin_classes, datetime.now())

        self.solver_initialized = False
        self.model_initialized = True

    def initializeSolver(self, solver:Solvers.Solver, propensity_caching:bool = True, no_rules_behaviour:str = "step"):
        if self.model_initialized:
            self.no_rules_behaviour = no_rules_behaviour
            self.propensity_caching = propensity_caching
            if propensity_caching:
                self.rule_propensity_update_dict = RuleChain.returnOneStepRuleUpdates(self.rules, self.locations, self.matched_indices, self.model_state.returnModelClasses())
            else:
                self.rule_propensity_update_dict = {}
            self.solver = solver(self.locations, self.rules, self.matched_indices, self.model_state, propensity_caching, no_rules_behaviour, self.rule_propensity_update_dict)
            self.solver_initialized = True
        else:
            raise(ValueError("Model not initialized: initialize model before solver"))

    def resetSimulation(self):
        for location in self.locations:
            location.reset()
        # Trajectory uses current location values so needs to be defined after location values reset.
        self.trajectory = Trajectory.Trajectory(self.locations)
        self.model_state.reset()
        self.solver.reset()

    def simulate(self, start_date, time_limit, max_iterations:int = 1000):
        self.start_date = start_date
        self.model_state = State.ModelState(self.builtin_classes, self.start_date)

        if self.solver_initialized and self.model_initialized:
            self.resetSimulation()
            start_perf_time = time.perf_counter()
            while self.model_state.elapsed_time < time_limit and self.model_state.iterations < max_iterations:
                # Simulate one step should update the location objects automatically with the new compartment values.
                new_time = self.solver.simulateOneStep(self.model_state.elapsed_time)
                self.model_state.processUpdate(new_time)
                # TODO To save memory - could just add changed location values
                for location_index, location in enumerate(self.locations):
                    self.trajectory.addEntry(new_time, location.class_values, location_index)

                if new_time is None:
                    break
            end_perf_time = time.perf_counter()
            print(f"The simulation has finished after {self.model_state.elapsed_time} {self.model_state.time_measurement}, requiring {self.model_state.iterations} iterations and {end_perf_time-start_perf_time} secs of compute time")
            return self.trajectory
        else:
            raise(ValueError("Model/solver not initialized: initialize model before the solver."))
